//--------------------------------------------
//  Template Class:  BaseArray<T>
//
//     Simplified array template class, 
//     with dynamic memory handling. Can
//     be used as base for any (direct)
//     one-dimensional array in C++.
//--------------------------------------------
//               Harris Georgiou, 1996.
//--------------------------------------------



#define         TCOUNT          int


#ifndef     __ABSTRACT_ARRAY_H
#define     __ABSTRACT_ARRAY_H

template <class T> class AbstractArray
{
    // We can declare only general member functions
    // that is functions that do not contain class' name

    public:
        virtual int  IsVoid( void )=0;
        virtual int  IsValid( void )=0;

        virtual TCOUNT  Alloc( TCOUNT size )=0;
        virtual void Init( T const &initVal, TCOUNT startFrom=0 )=0;
        virtual void Init( T const *initVal, TCOUNT startFrom=0 )=0;
        virtual void Done( void )=0;
        virtual TCOUNT  GetSize( void )=0;
        virtual TCOUNT  Resize( TCOUNT newsize )=0;
        virtual T &  ItemAt( TCOUNT idx )=0;
    
        virtual operator int()=0;
        virtual T & operator []( TCOUNT idx )=0;
        virtual T & operator ()( TCOUNT idx )=0;
    
};


#endif


#ifndef     __BASEARRAYS_H
#define     __BASEARRAYS_H

//#include <_null.h>
#ifndef     NULL
#define     NULL    0
#endif
//#include "_func.cpp"


template <class T> class BaseArray : public AbstractArray<T>
{
    protected:
        T       *arr;
        TCOUNT     sz;

    public:
        BaseArray()                     {  arr=NULL;  sz=(TCOUNT)0;  }
        BaseArray( TCOUNT size )        {  arr=NULL;  sz=(TCOUNT)0;  Resize(size);  }
        BaseArray( BaseArray &a )       {  arr=NULL;  sz=(TCOUNT)0;  Copy(a);  }

        ~BaseArray()                    {  if (arr) delete [] arr;  }

        int  IsVoid( void )             {  return(arr==NULL);  }
        int  IsValid( void )            {  return((arr!=NULL)&&(sz>(TCOUNT)0));  }

        TCOUNT  Alloc( TCOUNT size );
        void Init( const T &initVal, TCOUNT startFrom=0 );
        void Init( const T *initVal, TCOUNT startFrom=0 );
        void Done( void )               {  if (arr) delete [] arr;  sz=(TCOUNT)0;  }
        TCOUNT  Copy( const BaseArray &a );
        TCOUNT  Copy( const BaseArray *a );
        TCOUNT  GetSize( void )            {  return(sz);  }
        TCOUNT  Resize( TCOUNT newsize );
        T &  ItemAt( TCOUNT idx )          {  return(arr[idx]);  }
    
        operator int()                  {  return(IsValid());  }
        BaseArray & operator =( BaseArray &a )          {  Copy(a); return(a);  }
        BaseArray * operator =( BaseArray *a )          {  Copy(a); return(a);  }
        T & operator []( TCOUNT idx )      {  return(ItemAt(idx));  }
        T & operator ()( TCOUNT idx )      {  return(ItemAt(idx));  }
};



template <class T>
TCOUNT BaseArray<T>::Alloc( TCOUNT size )
{
    if ((size>(TCOUNT)0)&&(IsVoid()))
    {  arr=new T[size];
       if (arr)  sz=size;  else  sz=(TCOUNT)0;  }
    return(sz);  
}


template <class T>
void BaseArray<T>::Init( const T &initVal, TCOUNT startFrom=0 )
{
    if (IsValid())
    for (TCOUNT c=((startFrom>=(TCOUNT)0)?startFrom:(TCOUNT)0); c<sz; c++ )
        arr[c]=initVal;
}


template <class T>
void BaseArray<T>::Init( const T *initVal, TCOUNT startFrom=0 )
{
    if (IsValid())
    for (TCOUNT c=((startFrom>=(TCOUNT)0)?startFrom:(TCOUNT)0); c<sz; c++ )
        memcpy(&(arr[c]),initVal,sizeof(T));
}


template <class T>
TCOUNT BaseArray<T>::Copy( const BaseArray<T> &a )
{
    if (sz!=a.GetSize())
       {  Done();  Alloc(a.GetSize());  }
       
    if (sz!=a.GetSize())  return(-1);
    else  {
    for (TCOUNT c=(TCOUNT)0; c<a.GetSize(); c++)  arr[c]=a.ItemAt(c);
    return(sz);  }
}


template <class T>
TCOUNT BaseArray<T>::Copy( const BaseArray<T> *a )
{
    if (sz!=a.GetSize())
       {  Done();  Alloc(a.GetSize());  }
       
    if (sz!=a.GetSize())  return(-1);
    else  {
    for (TCOUNT c=0; c<a.GetSize(); c++)  memcpy(&(arr[c]),&(a.ItemAt(c)),sizeof(T));
    return(sz);  }
}


template <class T>
TCOUNT BaseArray<T>::Resize( TCOUNT newSize )
{
    T   *newarr;


    if (IsVoid())  return(Alloc(newSize));
    else
    {
        if ((newSize>0)&&(newSize!=sz))
        {  
            // The following alternative does not work (Null pointer assignment)
            // if (realloc(arr,newSize*sizeof(T))!=NULL)   sz=newSize;

            newarr = new T[newSize];
            if (newarr!=NULL)
            {
                for ( TCOUNT i=(TCOUNT)0; (i<newSize)&&(i<sz); i++ )
                    newarr[i] = arr[i];

                Done(); 
                arr = newarr;
                sz = newSize;
            }
        }
        
        return(sz);
    }
}


#endif

#define     _STANDALONE     1


#ifdef      _STANDALONE

#include <stdio.h>

void main( void )
{
    TCOUNT    c;
    BaseArray<int>  ia;

    ia.Resize(10);
    ia.Init(-1,2);
    printf("Initialized array (sz=%d)\n",ia.GetSize());
    for ( c=0; c<ia.GetSize(); c++ )  printf("%d  ",ia[c]);
    printf("\n");
    printf("Modified array (sz=%d)\n",ia.GetSize());
    for ( c=0; c<ia.GetSize(); c++ )  { ia[c]=c; printf("%d  ",ia[c]); }
    printf("\n");
    ia.Resize(ia.GetSize()+5);
    printf("Resized array (sz=%d)\n",ia.GetSize());
    for ( c=0; c<ia.GetSize(); c++ )  printf("%d  ",ia[c]);
}

#endif